#!/usr/bin/python3

import sqlite3

codes = {
          0: ( 2,  8, '  ld b,h\n  ld c,l\n'),
          1: ( 1, 11, '  add hl,hl\n'),
          2: ( 1, 11, '  add hl,bc\n'),
          3: ( 3, 19, '  or a\n  sbc hl,bc\n'),
          4: ( 6, 24, '  xor a\n  sub l\n  ld l,a\n  sbc a,a\n  sub h\n  ld h,a\n'),
         # 64: (14, 55, '  ld a,h\n  ld h,l\n  ld l,0\n  rra\n  rr h\n  rr l\n  rra\n  rr h\n  rr l\n'),
         64: (12, 47, '  ld a,h\n  ld h,l\n  rra\n  rr h\n  rra\n  rr h\n  rra\n  and 11000000b\n  ld l,a\n'),
        128: ( 9, 35, '  ld a,h\n  ld h,l\n  ld l,0\n  rra\n  rr h\n  rr l\n'),
        256: ( 3, 11, '  ld h,l\n  ld l,0\n'),
        255: ( 7, 28, '  xor a\n  sub l\n  ld l,a\n  add a,h\n  cpl\n  inc a\n  ld h,a\n'),
        257: ( 3, 12, '  ld a,h\n  add a,l\n  ld h,a\n'),
        258: ( 4, 23, '  ld a,l\n  add hl,hl\n  add a,h \n  ld h,a\n'),
        260: ( 5, 34, '  ld a,l\n  add hl,hl\n  add hl,hl\n  add a,h\n  ld h,a\n'),
        264: ( 6, 45, '  ld a,l\n  add hl,hl\n  add hl,hl\n  add hl,hl\n  add a,h \n  ld h,a\n'),
        513: ( 6, 37, '  ex de,hl\n  ld h,e\n  ld l,0\n  add hl,hl\n  add hl,de\n'),
        768: ( 6, 23, '  ld a,l\n  add a,a\n  add a,l\n  ld h,a\n  ld l,0\n'),
      32768: ( 6, 24, '  xor a\n  rr l\n  ld l,a\n  rra\n  ld h,a\n'),
  65536+255: ( 6, 31, '  ex de,hl\n  ld h,e\n  xor a\n  ld l,a\n  sbc hl,de\n'),
        }

speed_opt ={
              1 : (0, 0, '', ''),
              2 : (1, 11, 'B', codes[2][2]),
              3 : (4, 30, 'ABC', codes[3][2]),
             64 : ( codes[64][0],  codes[64][1], 'Z00064',  codes[64][2]),
            128 : (codes[128][0], codes[128][1], 'Z00128', codes[128][2]),
            255 : (codes[255][0], codes[255][1], 'Z00255', codes[255][2]),
            256 : (codes[256][0], codes[256][1], 'Z00256', codes[256][2]),
            257 : (codes[257][0], codes[257][1], 'Z00257', codes[257][2]),
            258 : (codes[258][0], codes[258][1], 'Z00258', codes[258][2]),
            260 : (codes[260][0], codes[260][1], 'Z00260', codes[260][2]),
            264 : (codes[264][0], codes[264][1], 'Z00264', codes[264][2]),
            513 : (codes[513][0], codes[513][1], 'Z00513', codes[513][2]),
            768 : (codes[768][0], codes[768][1], 'Z00768', codes[768][2]),
          32768 : (codes[32768][0], codes[32768][1], 'Z32768', codes[32768][2])
           }
size_opt = speed_opt.copy()
size_opt[255] = (codes[65536+255][0], codes[65536+255][1], 'Y00255', codes[65536+255][2])


def eval(code, do_speed=False):
    if 'ABD' in code or 'ABA' in code or 'DBC' in code or 'CBD' in code:
        return
    code = code.replace('BBBBBBBBBBBBBBB', 'Z32768').replace('BBBBBBBB', 'Z00256')
    if do_speed:
        code = code.replace('Z0256D', 'Z00255')
    x = 1
    y = 0
    size = 0
    speed = 0
    out = ''
    k = 0
    while k < len(code):
        i = ord(code[k]) - ord('A')
        k += 1
        if i in (24, 25): # 'Y', 'Z'
            i = ((25-i)<<16) + int(code[k:k+5])
            k += 5

        if i not in (1, 3):
            size  += codes[i][0]
            speed += codes[i][1]
            out   += codes[i][2]
            if i >= 8:
                x = x * i
        if i == 0:
            y = x
        elif i == 1:
            x += x
            if do_speed and x%256==0:
                out += '  sla h\n'
                size += 2
                speed += 8
            else:
                size  += codes[i][0]
                speed += codes[i][1]
                out   += codes[i][2]
        elif i == 2:
            x += y
        elif i == 3:
            x -= y
            if y%256 == 0:
                size += 3
                speed += 12
                out += '  ld a,h\n  sub b\n  ld h,a\n'
            else:
                size  += codes[i][0]
                speed += codes[i][1]
                out   += codes[i][2]
        elif i == 4:
            x = 65536 - x
    return (x, y, size, speed, out)

def upd(x, code, do_speed=False):
    global speed_opt, size_opt
    (x, y, size, speed, out) = eval(code, do_speed)
    code = code.replace('BBBBBBBBBBBBBBB', 'Z32768').replace('BBBBBBBB', 'Z00256')
    # if x == 33154/2:
    #     print(size, speed, code)
    if x in speed_opt:
        # Need to check which is better
        if size_opt[x][0] > size or (size_opt[x][0] == size and size_opt[x][1] > speed):
            size_opt[x] = (size, speed, code, out)
        if speed_opt[x][1] > speed or (speed_opt[x][1] == speed and speed_opt[x][0] > size):
            speed_opt[x] = (size, speed, code, out)
    else:
        speed_opt[x] = (size, speed, code, out)
        size_opt[x] = (size, speed, code, out)
    if 'BBBBBBB' in code:
        upd(x, code.replace('BBBBBBB', 'Z00128'))
    elif 'BBBBBB' in code:
        upd(x, code.replace('BBBBBB', 'Z00064'))
    if not do_speed:
        upd(x, code, True)
def mul(x):
    code = ''
    k = x

    if k in speed_opt:
        return

    # Let's check real quick if negating will be more efficient
    if k > 32768:
        upd(k, speed_opt[65536 - k][2] + 'E')
        upd(k, size_opt[65536 - k][2] + 'E')

    if k%2 == 0:
        upd(x, speed_opt[k/2][2] + 'B')
        upd(x, size_opt[k/2][2] + 'B')
        return

    # If me make it here, k == x, and it is odd

    # We'll first generate some code using the add-and-shift method
    odd = k
    while k > 1:
        if k%16 == 15:
            k += 1
            code = 'D' + code
        else:
            k -= 1
            code = 'C' + code
        while k%2 == 0:
            code = 'B' + code
            k /= 2
    if x == 16577:
        print('here:')
    upd(x, 'A' + code)

    # Now we'll calculate all of the factors and use that to see if is is more
    # efficient to generate the product based on its factors. (i.e. maybe it is
    # faster to multiply by 15 by multilying by 3, then 5)
    k = 3
    while k < 256:
        if odd%k == 0:
            upd(x, speed_opt[k][2] + speed_opt[odd/k][2])
            upd(x, size_opt[k][2] + size_opt[odd/k][2])
        k += 2
    return

#
# ld b,h
# ld c,l
# add hl,hl ;2
# add hl,hl ;4
# add hl,hl ;8
# add hl,hl ;16
# add hl,hl ;32
# add hl,hl ;64
# add hl,hl ;128
# add hl,bc ;129
# add hl,hl ;258
# add hl,bc ;259
# add hl,hl ;518
# add hl,hl ;1036
# add hl,hl ;2072
# add hl,hl ;4144
# add hl,hl ;8288
# add hl,hl ;16576
# add hl,bc ;16577
# add hl,hl ;33154

for i in range(1, 65536):
    mul(i)
    if size_opt[i][0] != speed_opt[i][0] or size_opt[i][1] != speed_opt[i][1]:
        (size, speed, code, out) = size_opt[i]
        print('mul_hl_{}:  ; size optimized'.format(i))
        # print(';' + size_opt[i][2])
        print(';{} bytes, {}cc'.format(size, speed))
        print(out)

        (size, speed, code, out) = speed_opt[i]
        print('mul_hl_{}:  ; speed optimized'.format(i))
        # print(';' + speed_opt[i][2])
        print(';{} bytes, {}cc'.format(size, speed))
        print(out)
    else:
        (size, speed, code, out) = size_opt[i]
        print('mul_hl_{}:'.format(i))
        # print(';' + size_opt[i][2])
        print(';{} bytes, {}cc'.format(size, speed))
        print(out)
#print(z80mul(256))
